import java.awt.*;
import java.awt.event.*;
import java.beans.*;
import javax.swing.*;
import javax.swing.table.*;
import javax.swing.event.*;
import java.util.*;
import java.sql.*;

/**
 * The DataGatherer is the base for all the data collection.  It lets subclasses get the
 * data, and then it knows how to process it and send it to the SegmentedData data models.
 *
 * @author tjacopi
 *
 */
public abstract class DataGatherer implements ActionListener {

    protected static final String LatencyPropertyKey = "Latency_";
    protected static final String ThroughputPropertyKey = "Throughput_";
    public    static final String NoDataMsg             = "No_Data";
    public    static final String DataStartMsg             = "Start_Data";
    public    static final String DataEndMsg             = "End_Data";

    protected javax.swing.Timer   timer           = null;
    protected Connection conn = null;
    protected PreparedStatement[] savedStmts = null;
    protected Hashtable  lastTimes = new Hashtable();
    protected Timestamp[]  lastRowFetched = null;
    protected String pwd = null;
    protected String schema = null;
    protected PropertyChangeSupport pcs = null;

    public DataGatherer(String schema) {
      this.schema = schema;
      pcs = new PropertyChangeSupport(this);
    }

    public boolean isRunning() {
      return (timer != null && timer.isRunning() );
    }

    public int getTimerDelay() {
      int delay = -1;
      if (isRunning()) {
        delay = timer.getDelay();
      };
      return delay;
    }

    public String getLatencyPropertyKeyFor(String recvq) {
      return LatencyPropertyKey + recvq;
    }

    public String getThroughputPropertyKeyFor(String recvq) {
      return ThroughputPropertyKey + recvq;
    }

    public void addPropertyChangeListener(PropertyChangeListener listener) {
      pcs.addPropertyChangeListener(listener);
    }

    public void removePropertyChangeListener(PropertyChangeListener listener) {
      pcs.removePropertyChangeListener(listener);
    }

    public void start(Connection connection) {
      this.conn = connection;
      int refreshInterval = getRefreshInterval();

      // Start a timer to fetch updates.
      timer = new javax.swing.Timer(refreshInterval, this );
      timer.setInitialDelay(0);                             // Tell timer to start now (no reason to wait for the interval to start)
      timer.start();
    }

    public void stop() {
      conn = null;
      try {
        if (timer != null) {
          timer.stop();
        };
        releaseStatements();
      } catch( Exception e ) {
        System.out.println(e);
        e.printStackTrace();
      }
    }



	/**
    * This is called whenever the timer pops.
	 */
    public void actionPerformed(ActionEvent evt) {
      Hashtable latestTimes = new Hashtable();                // Holds all the lastest time, by queues
      Hashtable latencyValues = new Hashtable();              // Holds all the latency times, by queues
      Hashtable latencyRowCounts = new Hashtable();           // Holds the number of latency rows found, by queues
      Hashtable tpValues = new Hashtable();                   // Holds all total row throughput, by queues
      try {
        if (savedStmts == null) {
          buildStatements();
        };

        pcs.firePropertyChange(DataStartMsg, null, null);     // Send this to notify people that new data is coming

        for (int i=0; i<savedStmts.length; i++) {
          savedStmts[i].setTimestamp(1, lastRowFetched[i] );
          ResultSet rs = savedStmts[i].executeQuery();

          // This will process the resultSet and fill out the hashtables
          lastRowFetched[i] = getValues(i, lastRowFetched[i], rs, latencyValues, latencyRowCounts, tpValues, latestTimes);

          rs.close();
        }


        // The latency data requires some post processing....do that now..
        for (Enumeration e = latencyValues.keys() ; e.hasMoreElements() ;) {
          String queue = (String) e.nextElement();
          double[] latencyValue = (double[]) latencyValues.get(queue);
          int[] latencyRowCount = (int[]) latencyRowCounts.get(queue);
          int numLatencyValues = latencyValue.length;
          if (latencyRowCount != null && latencyRowCount[0] > 1) {         // If more than one row was found
            for (int i=numLatencyValues-1; i>=0; i--) {
              latencyValue[i] = latencyValue[i]/(double) latencyRowCount[0]; // Divide to take the average of the rows
            }
          }
          for (int i=numLatencyValues-1; i>=0; i--) {
            latencyValue[i] = latencyValue[i]/1000.00;                      // Divide so value in seconds
          }

          // Now fire the notificaiton for latency on this queue
          String notificationKey = LatencyPropertyKey + queue;
          PropertyChangeEvent pce = new PropertyChangeEvent(this, notificationKey, null, latencyValue);
          pcs.firePropertyChange(pce);
        }

        for (Enumeration e = tpValues.keys() ; e.hasMoreElements() ;) {
          String queue = (String) e.nextElement();

          // Lets work on the throughput values now.
          Timestamp lastTime = (Timestamp) lastTimes.get(queue);  // Only can work if we have previous time
          if (lastTime != null) {
            Timestamp latestTime = (Timestamp) latestTimes.get(queue);
            long timeInMillisBetweenIntervals = latestTime.getTime() - lastTime.getTime();
            double intervalInSeconds = timeInMillisBetweenIntervals/1000.000;
            double[] rowTPValues = (double[]) tpValues.get(queue);
            rowTPValues[0] = rowTPValues[0]/intervalInSeconds;
            String notificationKey = ThroughputPropertyKey + queue;
            PropertyChangeEvent pce = new PropertyChangeEvent(this, notificationKey, null, rowTPValues);
            pcs.firePropertyChange(pce);
          };

        }                                          // end for loop

        for (Enumeration e = latestTimes.keys() ; e.hasMoreElements() ;) {
          String queue = (String) e.nextElement();
          lastTimes.put(queue, latestTimes.get(queue) );
        }

        // If we fetched no data, inform our listeners
        if (latencyValues.size() == 0) {
          pcs.firePropertyChange(NoDataMsg, null, null);
        };

        pcs.firePropertyChange(DataEndMsg, null, null);     // Send this to notify people that all data has been sent coming

      } catch( Exception e ) {
        System.out.println(e);
        e.printStackTrace();
      }
    }

   protected void buildStatements() throws SQLException {
     String[] queries = getMonitorQueries();
     savedStmts = new PreparedStatement[queries.length];
     lastRowFetched = new Timestamp[queries.length];
     int refreshInterval = timer.getDelay();
     for (int i=0; i<savedStmts.length; i++) {
       savedStmts[i] = conn.prepareStatement(queries[i]);
       lastRowFetched[i] = new Timestamp(System.currentTimeMillis() - refreshInterval );  // Set up start time so we can fetch the last interval
     }
   }

   protected void releaseStatements() {
     if (savedStmts != null) {
       for (int i=0; i<savedStmts.length; i++) {
         if (savedStmts[i] != null) {
           try {
             savedStmts[i].close();
           } catch(Throwable t) {
             System.out.println("Caught error when trying to close statements " + t);
             t.printStackTrace();
           }

           savedStmts[i] = null;
         };
       }
       savedStmts = null;
     }
   }

   protected int getRefreshInterval() {
      int updateTime = 5000;

      try {
        String sql = getRefreshIntervalQuery();
        PreparedStatement stmt = conn.prepareStatement(sql);
        ResultSet rs = stmt.executeQuery();
        rs.next();
        updateTime = rs.getInt(1);
        String catalogLevel = rs.getString(2);
        if (updateTime <= 0 ) {                   // If subsecond monitoring
          updateTime = 1;                        // ..just use 1 sec so graph does not move too fast
        }
        if (catalogLevel != null && catalogLevel.startsWith("08") ) {
          updateTime = updateTime * 1000;              // get time in millis
        };
        rs.close();
        stmt.close();
      } catch( Exception e ) {
        System.out.println(e);
        e.printStackTrace();
      }

      // Insure that we do not update the gui any more that once a second.
      // More than that is too fast....
      if (updateTime < 1000) {
        updateTime = 1000;
      };
      System.out.println("Using refresh value of " + updateTime + " milliseconds" );
      return updateTime;
   }

   protected abstract String[] getMonitorQueries();
   protected abstract String getRefreshIntervalQuery();
   protected abstract Timestamp getValues(int queryNumber, Timestamp previousLatestMonitorTime, ResultSet rs, Hashtable latencyValues, Hashtable latencyRowCounts, Hashtable tpValues, Hashtable latestTimes);
}