import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import java.util.*;
import java.text.*;
import java.beans.*;

/**
 * The visual graph control.
 *
 * @author tjacopi
 *
 */
public class ScrollingGraph extends JComponent implements PropertyChangeListener {

  protected static final int pixelBorder = 3;
  protected static final int barWidth    = 3;
  protected static final int barSpace    = 3;

  protected DecimalFormat decimalFormat = new DecimalFormat("#,###,##0.0");

  protected double    maxLabelValue  = 0.0;
  protected String    labelText      = null;
  protected Rectangle bounds = new Rectangle();
  protected Insets    insets = new Insets(0,0,0,0);
  protected Rectangle savedBounds = null;
  protected Rectangle graphRect = null;
  protected Rectangle labelRect = null;
  protected double    maxValue  = 0.0;
  protected int       maxLabelWidth = 0;
  protected int       labelHeight   = 0;
  protected int       fontAscent    = 0;
  protected int       maxBarHeight  = 0;

  protected SegmentedData data = null;
  protected Color[]  segmentColors = null;
  protected Color    extendedValueColor = null;

    public ScrollingGraph(Color[] colorEntries, SegmentedData data, String labelText, double maxLabelValue) {
      this.labelText = labelText;
      this.maxLabelValue = maxLabelValue;
      this.data = data;
      if (colorEntries == null || colorEntries.length == 0) {
        colorEntries = new Color[1];
        colorEntries[0] = Color.cyan;
      };
      segmentColors = colorEntries;
      setPreferredSize( new Dimension(280,60) );
      data.addPropertyChangeListener(this);
    }

    public ScrollingGraph(SegmentedData data, String labelText, double maxLabelValue) {
      this(null, data, labelText, maxLabelValue);
    }

    public void displayNonZeroExtendedValuesAs(Color clr) {
      extendedValueColor = clr;
    }

    public void reset() {
      if (maxValue != 0.0) {
        maxValue  = 0.0;
        repaint();
      };
    }

    protected void paintComponent(Graphics g) {
      getBounds(bounds);
      getInsets(insets);
      bounds.x = bounds.x + insets.left;
      bounds.y = bounds.y + insets.top;
      bounds.width = bounds.width - insets.right - insets.left;
      bounds.height = bounds.height - insets.top - insets.bottom;

      Dimension d = getSize();

      if (maxLabelWidth == 0) {
        maxLabelWidth = getPreferredLabelWidth();
      };
      if (labelHeight == 0) {
        FontMetrics fm = getFontMetrics( getFont() );
        labelHeight = fm.getHeight();
        fontAscent  = fm.getAscent();
      };

      if (savedBounds == null || !savedBounds.equals(bounds) ) {

        labelRect = new Rectangle(bounds.width - pixelBorder - maxLabelWidth,
                                  pixelBorder,
                                  maxLabelWidth,
                                  bounds.height - (2*pixelBorder) );

        graphRect = new Rectangle(pixelBorder,
                                  pixelBorder,
                                  labelRect.x - (2*pixelBorder),
                                  bounds.height - (2*pixelBorder) );

        maxBarHeight = graphRect.height - 1 - (labelHeight/2);   // -1 for line at bottom
        savedBounds = (Rectangle) bounds.clone();
//      System.out.println("insets is " + insets);
//      System.out.println("bounds is " + bounds);
//      System.out.println("labelRect is " + labelRect);
//      System.out.println("graphRect is " + graphRect);
//      System.out.println("maxbarheight is " + maxBarHeight + "  fontAscent " + fontAscent + " labelHeight " + labelHeight);
      }

      g.setColor(Color.black);
      g.fillRect(0, 0, bounds.width, bounds.height);

      // Calc the top of the bar graph
      int numBars = ((graphRect.width-barWidth-1)/(barWidth+barSpace)) + 1;    // How many bars are visible, +1 for rounding
      double currentVisibleMaxValue = data.getMaxValue(numBars);
//    System.out.println("calc numBars is " +numBars);
      if (   currentVisibleMaxValue >= maxValue  ||             // If we have a new high value
           ((currentVisibleMaxValue/maxValue) < 0.6 ) ) {       // ..or if the showing values are not 60% of the high
        maxValue = currentVisibleMaxValue;                      // ..then use the current high as the high
      };

      paintLabels(g);

      paintGraph(g);
    }

    protected void paintGraph(Graphics g) {
      int numSegments  = data.getNumSegments();
      int actualValues = data.getNumValues();
      int bottomGraph = graphRect.y + graphRect.height -1;
      int topGraph    = graphRect.y + labelHeight/2;
      g.setColor(Color.cyan.darker() );
      g.drawLine(graphRect.x, graphRect.y + graphRect.height, graphRect.width, graphRect.y+graphRect.height);
//    System.out.println("drew baseline at " + graphRect.y + graphRect.height);
      g.setColor(Color.cyan.darker().darker() );
      g.drawLine(graphRect.x, topGraph, graphRect.width, topGraph);
      g.drawLine(graphRect.x, topGraph + (maxBarHeight/2), graphRect.width, topGraph + (maxBarHeight/2));

      int barXPos = graphRect.x + graphRect.width - 1 - barWidth;
      int numBars = 1;
      while(barXPos >0 && numBars<=actualValues) {

        double totalRowValue = 0.0;
        double[] rowValues = data.getDataValues(numBars-1);
        for (int j=0; j<numSegments; j++) {
          totalRowValue = totalRowValue + rowValues[j];
        }
        if (totalRowValue>0.0) {
          int totalBarHeight = (int) ((totalRowValue/maxValue) * maxBarHeight);
          int segmentYPos = graphRect.y + graphRect.height - totalBarHeight;
//        System.out.println("painting row, numSegments = " + numSegments + " totalBarHeight " + totalBarHeight + " startingYPos " + segmentYPos);
          for (int j=0; j<numSegments-1; j++) {
//          g.setColor( segmentColors[j] );
            g.setColor( getColorForSegment(j, numBars-1) );
            double segmentPct = rowValues[j]/totalRowValue;
            int segmentHeight = (int) (segmentPct * totalBarHeight);
//          System.out.println("filling bar from " + segmentYPos + " to " + segmentHeight + " with color " + segmentColors[j]);
            g.fillRect(barXPos, segmentYPos, barWidth, segmentHeight);
            segmentYPos = segmentYPos + segmentHeight;
          }

          // For the last segment, paint all thats left so no spaces due to rounding errors
//        g.setColor( segmentColors[numSegments-1] );
          g.setColor( getColorForSegment(numSegments-1, numBars-1) );
          int distanceLeft = (graphRect.y + graphRect.height ) - segmentYPos;
//        System.out.println("filling last bar from " + segmentYPos + " to " + distanceLeft + " with color " + segmentColors[numSegments-1]);
          g.fillRect(barXPos, segmentYPos, barWidth, distanceLeft);
        };
        numBars++;
        barXPos = barXPos - barWidth - barSpace;
      }
    }

    protected void paintLabels(Graphics g) {
      g.setColor(Color.yellow);
      double halfValue = maxValue/2.0;
      Object[] parms = new Object[1];

      String txt = decimalFormat.format(maxValue);
      parms[0] = txt;
      String label = MessageFormat.format(labelText, parms);
      g.drawString(label, labelRect.x, labelRect.y+fontAscent);

      String txt2 = decimalFormat.format(halfValue);
      parms[0] = txt2;
      String label2 = MessageFormat.format(labelText, parms);
      g.drawString(label2, labelRect.x, labelRect.y+(maxBarHeight/2)+fontAscent);
    }

    protected Color getColorForSegment(int segmentIndex, int dataIndex) {
      Color clr = segmentColors[segmentIndex];
      if (extendedValueColor != null) {
        if (data.getExtendedDataValues(dataIndex) != 0.0) {
          clr = extendedValueColor;
        };
      };

      return clr;
    }

    public void propertyChange(PropertyChangeEvent e) {
      repaint();
    }

    protected void setLabelWidth(int width) {
      maxLabelWidth = width;
    }

    protected int getPreferredLabelWidth() {
      FontMetrics fm = getFontMetrics( getFont() );
      String txt = decimalFormat.format(maxLabelValue);
      Object[] parms = new Object[1];
      parms[0] = txt;
      String maxLabelString = MessageFormat.format(labelText, parms);
      int labelWidth = fm.stringWidth(maxLabelString);
      return labelWidth;
    }
}