import javax.swing.*;
import java.util.*;
import java.beans.*;

/**
 * The data model used for the scrolling graph.
 *
 * @author tjacopi
 *
 */
public class SegmentedData implements PropertyChangeListener   {

  protected int       numSegments  = 0;
  protected int       totalValues = 0;
  protected volatile  double[][] dataValues = null;
  protected volatile  double[]   extendedDataValues = null;
  protected int      actualValues = 0;
  protected String    propertyKeyValue = null;
  protected double[]  noDataValue = null;
  protected volatile  boolean hadValueInInterval = false;

  protected PropertyChangeSupport pcs = null;


//public SegmentedData(int numSegments, int maxValues, String propertyKeyValue) {
//  this(numSegments, maxValues, propertyKeyValue, null);
//}

  /**
   *  @param  noDataValue  If no data was recieved for this interval, this value is then used as the
                           interval value.  If null, then no data is saved and the interval is ignored.
                           If the no data value was used, you can determine by calling "wasNoDataValueUsed(int)"
   */
  public SegmentedData(int numSegments, int maxValues, String propertyKeyValue, double[] noDataValue) {
    this.numSegments = numSegments;
    this.totalValues = maxValues;
    this.propertyKeyValue = propertyKeyValue;
    this.noDataValue = noDataValue;

    dataValues = new double[totalValues][numSegments];
    extendedDataValues = new double[totalValues];
    for (int i=0; i<totalValues; i++) {
      extendedDataValues[i] = 0.0;
      for (int j=0; j<numSegments; j++) {
        dataValues[i][j] = 0.0;
      }
    }

    pcs = new PropertyChangeSupport(this);
  }

    public void propertyChange(PropertyChangeEvent e) {
      if ( propertyKeyValue.equals(e.getPropertyName() ) ) {
        hadValueInInterval = true;                                            // Remember that we saw one
        double[] newValues = (double[]) e.getNewValue();
        SwingUtilities.invokeLater( new AddValuesRunnable(newValues) );       // Do this so on event thread
      } else if ( DataGatherer.DataStartMsg.equals(e.getPropertyName() ) ) {
        hadValueInInterval = false;                                           // The interval is starting...so reset flag
      } else if ( DataGatherer.DataEndMsg.equals(e.getPropertyName() ) ) {
        if (noDataValue != null && !hadValueInInterval) {                     // If we did not see a data value, and have one to use
          SwingUtilities.invokeLater( new AddValuesRunnable(noDataValue) );   // ..then make believe we saw a data value
        }
      };

    }

    // This is a cheap way to insure that the graphic components do not see values change when
    // in the middle of repainting.
    private void addValueInternal(double[] values) {
      synchronized (dataValues) {
        for (int i=totalValues-2; i>=0; i--) {
          extendedDataValues[i+1] = extendedDataValues[i];
          for (int j=0; j<numSegments; j++) {
            dataValues[i+1][j] = dataValues[i][j];
          }
        }
        for (int j=0; j<numSegments; j++) {
          dataValues[0][j] = values[j];
        }

        if (values.length > numSegments) {
          extendedDataValues[0] = values[numSegments];      // for now, just support one additional value
        } else {
          extendedDataValues[0] = 0.0;
        }

        if (actualValues < dataValues.length) {
          actualValues++;
        }

      }

      PropertyChangeEvent pce = new PropertyChangeEvent(this, "NewDataValue", null, null);
      pcs.firePropertyChange(pce);
    }

    public void reset() {
      synchronized (dataValues) {
        for (int i=0; i<totalValues; i++) {
          for (int j=0; j<numSegments; j++) {
            dataValues[i][j] = 0.0;
          }
        }
        actualValues = 0;
      }
    }


    public double getMaxValue() {
      return getMaxValue(0);
    }

    public double getMaxValue(int numValues) {
      double maxValue = 0.0;
      synchronized (dataValues) {
        int numberOfValues = totalValues;
        if (numValues>0 && numValues<totalValues) {
          numberOfValues = numValues;
        }
        for (int i=numberOfValues-1; i>=0; i--) {
          double totalRowValue = 0.0;
          for (int j=0; j<numSegments; j++) {
            totalRowValue = totalRowValue + dataValues[i][j];
          }
          if (totalRowValue > maxValue) {
            maxValue = totalRowValue;
          };
        }
      }
      return maxValue;
    }

    public double getAverageValue(int numValues) {
      double totalValue = 0.0;
      int    usedValues = 0;
      synchronized (dataValues) {
        if (numValues > actualValues) {
          numValues = actualValues;
        };

        for (int i=0; i<numValues; i++) {
          double totalRowValue = 0.0;
          for (int j=0; j<numSegments; j++) {
            totalRowValue = totalRowValue + dataValues[i][j];
          }
          if (totalRowValue > 0) {
            totalValue = totalValue + totalRowValue;
            usedValues++;
          };
        }
      }

      if (usedValues>1) {
        totalValue = totalValue/usedValues;
      };
      return totalValue;
    }

    public int getNumValues() {
      return actualValues;
    }

    public int getNumSegments() {
      return numSegments;
    }

    public double[] getDataValues(int index) {
      return dataValues[index];
    }

    public double[] getLatestDataValues() {
      return getDataValues(0);
    }

    public double getExtendedDataValues(int index) {
      return extendedDataValues[index];
    }

    public double getLatestExtendedDataValues() {
      return getExtendedDataValues(0);
    }

    public boolean isLatestValueNonZero() {
      boolean nonZero = false;
      if (actualValues > 0) {
        for (int i=0; i<numSegments && !nonZero; i++ ) {
          if ( (dataValues[0][i]) != 0.0) {                         //If any values non zero, set flag
            nonZero = true;
          };
        }
      }
      return nonZero;
    }

    public boolean dataReceivedForLastInterval() {
      return hadValueInInterval;
    }

    public void addPropertyChangeListener(PropertyChangeListener listener) {
      pcs.addPropertyChangeListener(listener);
    }

    public void removePropertyChangeListener(PropertyChangeListener listener) {
      pcs.removePropertyChangeListener(listener);
    }

    class AddValuesRunnable implements Runnable {
      double[] values = null;
      public AddValuesRunnable(double[] values) {
        this.values = values;
      }
      public void run() {
        addValueInternal(values);
        values = null;
      }
    }

}